@@ -19,9 +19,9 @@ module Distribution.Client.Install (
   ) where
 
 import Data.List
-         ( unfoldr, find, nub, sort )
+         ( unfoldr, nub, sort, (\\) )
 import Data.Maybe
-         ( isJust, fromMaybe )
+         ( isJust, fromMaybe, maybeToList )
 import Control.Exception as Exception
          ( handleJust )
 #if MIN_VERSION_base(4,0,0)
@@ -42,17 +42,19 @@ import System.Directory
 import System.FilePath
          ( (</>), (<.>), takeDirectory )
 import System.IO
-         ( openFile, IOMode(AppendMode) )
+         ( openFile, IOMode(AppendMode), stdout, hFlush )
 import System.IO.Error
          ( isDoesNotExistError, ioeGetFileName )
 
 import Distribution.Client.Targets
 import Distribution.Client.Dependency
+import Distribution.Client.Dependency.Types
+         ( Solver(..) )
 import Distribution.Client.FetchUtils
 import qualified Distribution.Client.Haddock as Haddock (regenerateHaddockIndex)
 -- import qualified Distribution.Client.Info as Info
 import Distribution.Client.IndexUtils as IndexUtils
-         ( getAvailablePackages, getInstalledPackages )
+         ( getSourcePackages, getInstalledPackages )
 import qualified Distribution.Client.InstallPlan as InstallPlan
 import Distribution.Client.InstallPlan (InstallPlan)
 import Distribution.Client.Setup
@@ -62,7 +64,7 @@ import Distribution.Client.Setup
 import Distribution.Client.Config
          ( defaultCabalDir )
 import Distribution.Client.Tar (extractTarGzFile)
-import Distribution.Client.Types as Available
+import Distribution.Client.Types as Source
 import Distribution.Client.BuildReports.Types
          ( ReportLevel(..) )
 import Distribution.Client.SetupWrapper
@@ -74,6 +76,7 @@ import qualified Distribution.Client.InstallSymlink as InstallSymlink
          ( symlinkBinaries )
 import qualified Distribution.Client.Win32SelfUpgrade as Win32SelfUpgrade
 import qualified Distribution.Client.World as World
+import qualified Distribution.InstalledPackageInfo as Installed
 import Paths_cabal_install (getBinDir)
 
 import Distribution.Simple.Compiler
@@ -81,14 +84,15 @@ import Distribution.Simple.Compiler
          , PackageDB(..), PackageDBStack )
 import Distribution.Simple.Program (ProgramConfiguration, defaultProgramConfiguration)
 import qualified Distribution.Simple.InstallDirs as InstallDirs
-import qualified Distribution.Client.PackageIndex as PackageIndex
-import Distribution.Client.PackageIndex (PackageIndex)
+import qualified Distribution.Simple.PackageIndex as PackageIndex
+import Distribution.Simple.PackageIndex (PackageIndex)
 import Distribution.Simple.Setup
-         ( haddockCommand, HaddockFlags(..), emptyHaddockFlags
+         ( haddockCommand, HaddockFlags(..)
          , buildCommand, BuildFlags(..), emptyBuildFlags
          , toFlag, fromFlag, fromFlagOrDefault, flagToMaybe )
 import qualified Distribution.Simple.Setup as Cabal
-         ( installCommand, InstallFlags(..), emptyInstallFlags )
+         ( installCommand, InstallFlags(..), emptyInstallFlags
+         , emptyTestFlags, testCommand )
 import Distribution.Simple.Utils
          ( rawSystemExit, comparing )
 import Distribution.Simple.InstallDirs as InstallDirs
@@ -97,16 +101,17 @@ import Distribution.Simple.InstallDirs as InstallDirs
 import Distribution.Package
          ( PackageIdentifier, packageName, packageVersion
          , Package(..), PackageFixedDeps(..)
-         , Dependency(..), thisPackageVersion )
+         , Dependency(..), thisPackageVersion, InstalledPackageId )
 import qualified Distribution.PackageDescription as PackageDescription
 import Distribution.PackageDescription
-         ( PackageDescription )
+         ( PackageDescription, GenericPackageDescription(..), Flag(..)
+         , FlagName(..), FlagAssignment )
 import Distribution.PackageDescription.Configuration
          ( finalizePackageDescription )
 import Distribution.Version
          ( Version, anyVersion, thisVersion )
 import Distribution.Simple.Utils as Utils
-         ( notice, info, debug, warn, die, intercalate, withTempDirectory )
+         ( notice, info, warn, die, intercalate, withTempDirectory )
 import Distribution.Client.Utils
          ( inDir, mergeBy, MergeResult(..) )
 import Distribution.System
@@ -114,7 +119,7 @@ import Distribution.System
 import Distribution.Text
          ( display )
 import Distribution.Verbosity as Verbosity
-         ( Verbosity, showForCabal, verbose )
+         ( Verbosity, showForCabal, verbose, deafening )
 import Distribution.Simple.BuildPaths ( exeExtension )
 
 --TODO:
@@ -145,13 +150,16 @@ install, upgrade
   -> ConfigFlags
   -> ConfigExFlags
   -> InstallFlags
+  -> HaddockFlags
   -> [UserTarget]
   -> IO ()
 install verbosity packageDBs repos comp conf
-  globalFlags configFlags configExFlags installFlags userTargets0 = do
+  globalFlags configFlags configExFlags installFlags haddockFlags userTargets0 = do
 
-    installed     <- getInstalledPackages verbosity comp packageDBs conf
-    availableDb   <- getAvailablePackages verbosity repos
+    installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
+    sourcePkgDb       <- getSourcePackages    verbosity repos
+
+    solver <- chooseSolver verbosity (fromFlag (configSolver  configExFlags)) (compilerId comp)
 
     let -- For install, if no target is given it means we use the
         -- current directory as the single target
@@ -159,31 +167,36 @@ install verbosity packageDBs repos comp conf
                     | otherwise         = userTargets0
 
     pkgSpecifiers <- resolveUserTargets verbosity
-                       globalFlags (packageIndex availableDb) userTargets
+                       (fromFlag $ globalWorldFile globalFlags)
+                       (packageIndex sourcePkgDb)
+                       userTargets
 
     notice verbosity "Resolving dependencies..."
     installPlan   <- foldProgress logMsg die return $
                        planPackages
-                         comp configFlags configExFlags installFlags
-                         installed availableDb pkgSpecifiers
+                         comp solver configFlags configExFlags installFlags
+                         installedPkgIndex sourcePkgDb pkgSpecifiers
 
-    printPlanMessages verbosity installed installPlan dryRun
+    checkPrintPlan verbosity installedPkgIndex installPlan installFlags pkgSpecifiers
 
     unless dryRun $ do
       installPlan' <- performInstallations verbosity
-                        context installed installPlan
+                        context installedPkgIndex installPlan
       postInstallActions verbosity context userTargets installPlan'
 
   where
     context :: InstallContext
     context = (packageDBs, repos, comp, conf,
-               globalFlags, configFlags, configExFlags, installFlags)
+               globalFlags, configFlags, configExFlags, installFlags, haddockFlags)
 
     dryRun      = fromFlag (installDryRun installFlags)
-    logMsg message rest = debug verbosity message >> rest
+    logMsg message rest = debugNoWrap message >> rest
+    -- Solver debug output really looks better without automatic
+    -- line wrapping. TODO: This should probably be moved into
+    -- the utilities module.
+    debugNoWrap xs = when (verbosity >= deafening) (putStrLn xs >> hFlush stdout)
 
-
-upgrade _ _ _ _ _ _ _ _ _ _ = die $
+upgrade _ _ _ _ _ _ _ _ _ _ _ = die $
     "Use the 'cabal install' command instead of 'cabal upgrade'.\n"
  ++ "You can install the latest version of a package using 'cabal install'. "
  ++ "The 'cabal upgrade' command has been removed because people found it "
@@ -203,25 +216,28 @@ type InstallContext = ( PackageDBStack
                       , GlobalFlags
                       , ConfigFlags
                       , ConfigExFlags
-                      , InstallFlags )
+                      , InstallFlags
+                      , HaddockFlags )
 
 -- ------------------------------------------------------------
 -- * Installation planning
 -- ------------------------------------------------------------
 
 planPackages :: Compiler
+             -> Solver
              -> ConfigFlags
              -> ConfigExFlags
              -> InstallFlags
-             -> PackageIndex InstalledPackage
-             -> AvailablePackageDb
-             -> [PackageSpecifier AvailablePackage]
+             -> PackageIndex
+             -> SourcePackageDb
+             -> [PackageSpecifier SourcePackage]
              -> Progress String String InstallPlan
-planPackages comp configFlags configExFlags installFlags
-             installed availableDb pkgSpecifiers =
+planPackages comp solver configFlags configExFlags installFlags
+             installedPkgIndex sourcePkgDb pkgSpecifiers =
 
         resolveDependencies
           buildPlatform (compilerId comp)
+          solver
           resolverParams
 
     >>= if onlyDeps then adjustPlanOnlyDeps else return
@@ -229,7 +245,18 @@ planPackages comp configFlags configExFlags installFlags
   where
     resolverParams =
 
-        setPreferenceDefault (if upgradeDeps then PreferAllLatest
+        setMaxBackjumps (if maxBackjumps < 0 then Nothing
+                                             else Just maxBackjumps)
+
+      . setIndependentGoals independentGoals
+
+      . setReorderGoals reorderGoals
+
+      . setAvoidReinstalls avoidReinstalls
+
+      . setShadowPkgs shadowPkgs
+
+      . setPreferenceDefault (if upgradeDeps then PreferAllLatest
                                              else PreferLatestForSelected)
 
       . addPreferences
@@ -239,20 +266,30 @@ planPackages comp configFlags configExFlags installFlags
 
       . addConstraints
           -- version constraints from the config file or command line
-          [ PackageVersionConstraint name ver
-          | Dependency name ver <- configConstraints configFlags ]
+            (map userToPackageConstraint (configExConstraints configExFlags))
 
       . addConstraints
           --FIXME: this just applies all flags to all targets which
           -- is silly. We should check if the flags are appropriate
-          [ PackageFlagsConstraint (pkgSpecifierTarget pkgSpecifier) flags
+          [ PackageConstraintFlags (pkgSpecifierTarget pkgSpecifier) flags
           | let flags = configConfigurationsFlags configFlags
           , not (null flags)
           , pkgSpecifier <- pkgSpecifiers ]
 
+      . addConstraints
+          [ PackageConstraintStanzas (pkgSpecifierTarget pkgSpecifier) stanzas
+          | pkgSpecifier <- pkgSpecifiers ]
+
       . (if reinstall then reinstallTargets else id)
 
-      $ standardInstallPolicy installed availableDb pkgSpecifiers
+      $ standardInstallPolicy installedPkgIndex sourcePkgDb pkgSpecifiers
+
+    stanzas = concat
+        [ if testsEnabled then [TestStanzas] else []
+        , if benchmarksEnabled then [BenchStanzas] else []
+        ]
+    testsEnabled = fromFlagOrDefault False $ configTests configFlags
+    benchmarksEnabled = fromFlagOrDefault False $ configBenchmarks configFlags
 
     --TODO: this is a general feature and should be moved to D.C.Dependency
     -- Also, the InstallPlan.remove should return info more precise to the
@@ -260,11 +297,8 @@ planPackages comp configFlags configExFlags installFlags
     adjustPlanOnlyDeps :: InstallPlan -> Progress String String InstallPlan
     adjustPlanOnlyDeps =
         either (Fail . explain) Done
-      . InstallPlan.remove isTarget
+      . InstallPlan.remove (isTarget pkgSpecifiers)
       where
-        isTarget pkg = packageName pkg `elem` targetnames
-        targetnames  = map pkgSpecifierTarget pkgSpecifiers
-        
         explain :: [InstallPlan.PlanProblem] -> String
         explain problems =
             "Cannot select only the dependencies (as requested by the "
@@ -281,74 +315,210 @@ planPackages comp configFlags configExFlags installFlags
                   , depid <- depids
                   , packageName depid `elem` targetnames ]
 
-    reinstall   = fromFlag (installReinstall installFlags)
-    upgradeDeps = fromFlag (installUpgradeDeps installFlags)
-    onlyDeps    = fromFlag (installOnlyDeps installFlags)
+        targetnames  = map pkgSpecifierTarget pkgSpecifiers
+
+
+    reinstall        = fromFlag (installReinstall        installFlags)
+    reorderGoals     = fromFlag (installReorderGoals     installFlags)
+    independentGoals = fromFlag (installIndependentGoals installFlags)
+    avoidReinstalls  = fromFlag (installAvoidReinstalls  installFlags)
+    shadowPkgs       = fromFlag (installShadowPkgs       installFlags)
+    maxBackjumps     = fromFlag (installMaxBackjumps     installFlags)
+    upgradeDeps      = fromFlag (installUpgradeDeps      installFlags)
+    onlyDeps         = fromFlag (installOnlyDeps         installFlags)
 
 -- ------------------------------------------------------------
 -- * Informational messages
 -- ------------------------------------------------------------
 
-printPlanMessages :: Verbosity
-                  -> PackageIndex InstalledPackage
-                  -> InstallPlan
-                  -> Bool
-                  -> IO ()
-printPlanMessages verbosity installed installPlan dryRun = do
+-- | Perform post-solver checks of the install plan and print it if
+-- either requested or needed.
+checkPrintPlan :: Verbosity
+               -> PackageIndex
+               -> InstallPlan
+               -> InstallFlags
+               -> [PackageSpecifier SourcePackage]
+               -> IO ()
+checkPrintPlan verbosity installed installPlan installFlags pkgSpecifiers = do
 
+  -- User targets that are already installed.
+  let preExistingTargets =
+        [ p | let tgts = map pkgSpecifierTarget pkgSpecifiers,
+              InstallPlan.PreExisting p <- InstallPlan.toList installPlan,
+              packageName p `elem` tgts ]
+
+  -- If there's nothing to install, we print the already existing
+  -- target packages as an explanation.
   when nothingToInstall $
-    notice verbosity $
-         "No packages to be installed. All the requested packages are "
-      ++ "already installed.\n If you want to reinstall anyway then use "
-      ++ "the --reinstall flag."
+    notice verbosity $ unlines $
+         "All the requested packages are already installed:"
+       : map (display . packageId) preExistingTargets
+      ++ ["Use --reinstall if you want to reinstall anyway."]
 
-  when (dryRun || verbosity >= verbose) $
-    printDryRun verbosity installed installPlan
+  let lPlan = linearizeInstallPlan installed installPlan
+  -- Are any packages classified as reinstalls?
+  let reinstalledPkgs = concatMap (extractReinstalls . snd) lPlan
+  -- Packages that are already broken.
+  let oldBrokenPkgs =
+          map Installed.installedPackageId
+        . PackageIndex.reverseDependencyClosure installed
+        . map (Installed.installedPackageId . fst)
+        . PackageIndex.brokenPackages
+        $ installed
+  let excluded = reinstalledPkgs ++ oldBrokenPkgs
+  -- Packages that are reverse dependencies of replaced packages are very
+  -- likely to be broken. We exclude packages that are already broken.
+  let newBrokenPkgs =
+        filter (\ p -> not (Installed.installedPackageId p `elem` excluded))
+               (PackageIndex.reverseDependencyClosure installed reinstalledPkgs)
+  let containsReinstalls = not (null reinstalledPkgs)
+  let breaksPkgs         = not (null newBrokenPkgs)
+
+  let adaptedVerbosity
+        | containsReinstalls && not overrideReinstall = verbosity `max` verbose
+        | otherwise                                   = verbosity
+
+  -- We print the install plan if we are in a dry-run or if we are confronted
+  -- with a dangerous install plan.
+  when (dryRun || containsReinstalls && not overrideReinstall) $
+    printPlan (dryRun || breaksPkgs && not overrideReinstall) adaptedVerbosity lPlan
+
+  -- If the install plan is dangerous, we print various warning messages. In
+  -- particular, if we can see that packages are likely to be broken, we even
+  -- bail out (unless installation has been forced with --force-reinstalls).
+  when containsReinstalls $ do
+    if breaksPkgs
+      then do
+        (if dryRun || overrideReinstall then warn verbosity else die) $ unlines $
+            "The following packages are likely to be broken by the reinstalls:"
+          : map (display . Installed.sourcePackageId) newBrokenPkgs
+          ++ if overrideReinstall
+               then if dryRun then [] else
+                 ["Continuing even though the plan contains dangerous reinstalls."]
+               else
+                 ["Use --force-reinstalls if you want to install anyway."]
+      else unless dryRun $ warn verbosity
+             "Note that reinstalls are always dangerous. Continuing anyway..."
 
   where
     nothingToInstall = null (InstallPlan.ready installPlan)
 
+    dryRun            = fromFlag (installDryRun            installFlags)
+    overrideReinstall = fromFlag (installOverrideReinstall installFlags)
 
-printDryRun :: Verbosity
-            -> PackageIndex InstalledPackage
-            -> InstallPlan
-            -> IO ()
-printDryRun verbosity installed plan = case unfoldr next plan of
-  []   -> return ()
-  pkgs
-    | verbosity >= Verbosity.verbose -> notice verbosity $ unlines $
-        "In order, the following would be installed:"
-      : map showPkgAndReason pkgs
-    | otherwise -> notice verbosity $ unlines $
-        "In order, the following would be installed (use -v for more details):"
-      : map (display . packageId) pkgs
+linearizeInstallPlan :: PackageIndex
+                     -> InstallPlan
+                     -> [(ConfiguredPackage, PackageStatus)]
+linearizeInstallPlan installedPkgIndex plan = unfoldr next plan
   where
     next plan' = case InstallPlan.ready plan' of
       []      -> Nothing
-      (pkg:_) -> Just (pkg, InstallPlan.completed pkgid result plan')
-        where pkgid = packageId pkg
+      (pkg:_) -> Just ((pkg, status), InstallPlan.completed pkgid result plan')
+        where pkgid  = packageId pkg
+              status = packageStatus installedPkgIndex pkg
               result = BuildOk DocsNotTried TestsNotTried
               --FIXME: This is a bit of a hack,
               -- pretending that each package is installed
 
-    showPkgAndReason pkg' = display (packageId pkg') ++ " " ++
-          case PackageIndex.lookupPackageName installed (packageName pkg') of
-            [] -> "(new package)"
-            ps ->  case find ((==packageId pkg') . packageId) ps of
-              Nothing  -> "(new version)"
-              Just pkg -> "(reinstall)" ++ case changes pkg pkg' of
-                []   -> ""
-                diff -> " changes: "  ++ intercalate ", " diff
-    changes pkg pkg' = map change . filter changed
+data PackageStatus = NewPackage
+                   | NewVersion [Version]
+                   | Reinstall  [InstalledPackageId] [PackageChange]
+
+type PackageChange = MergeResult PackageIdentifier PackageIdentifier
+
+isTarget :: Package pkg => [PackageSpecifier SourcePackage] -> pkg -> Bool
+isTarget pkgSpecifiers pkg = packageName pkg `elem` targetnames
+  where
+    targetnames  = map pkgSpecifierTarget pkgSpecifiers
+
+extractReinstalls :: PackageStatus -> [InstalledPackageId]
+extractReinstalls (Reinstall ipids _) = ipids
+extractReinstalls _                   = []
+
+packageStatus :: PackageIndex -> ConfiguredPackage -> PackageStatus
+packageStatus installedPkgIndex cpkg =
+  case PackageIndex.lookupPackageName installedPkgIndex
+                                      (packageName cpkg) of
+    [] -> NewPackage
+    ps ->  case filter ((==packageId cpkg) . Installed.sourcePackageId) (concatMap snd ps) of
+      []           -> NewVersion (map fst ps)
+      pkgs@(pkg:_) -> Reinstall (map Installed.installedPackageId pkgs)
+                                (changes pkg cpkg)
+
+  where
+
+    changes :: Installed.InstalledPackageInfo
+            -> ConfiguredPackage
+            -> [MergeResult PackageIdentifier PackageIdentifier]
+    changes pkg pkg' = filter changed
                      $ mergeBy (comparing packageName)
-                         (nub . sort . depends $ pkg)
+                         -- get dependencies of installed package (convert to source pkg ids via index)
+                         (nub . sort . concatMap (maybeToList .
+                                                  fmap Installed.sourcePackageId .
+                                                  PackageIndex.lookupInstalledPackageId installedPkgIndex) .
+                                                  Installed.depends $ pkg)
+                         -- get dependencies of configured package
                          (nub . sort . depends $ pkg')
+
+    changed (InBoth    pkgid pkgid') = pkgid /= pkgid'
+    changed _                        = True
+
+printPlan :: Bool -- is dry run
+          -> Verbosity
+          -> [(ConfiguredPackage, PackageStatus)]
+          -> IO ()
+printPlan dryRun verbosity plan = case plan of
+  []   -> return ()
+  pkgs
+    | verbosity >= Verbosity.verbose -> notice verbosity $ unlines $
+        ("In order, the following " ++ wouldWill ++ " be installed:")
+      : map showPkgAndReason pkgs
+    | otherwise -> notice verbosity $ unlines $
+        ("In order, the following " ++ wouldWill ++ " be installed (use -v for more details):")
+      : map (display . packageId) (map fst pkgs)
+  where
+    wouldWill | dryRun    = "would"
+              | otherwise = "will"
+
+    showPkgAndReason (pkg', pr) = display (packageId pkg') ++
+          showFlagAssignment (nonDefaultFlags pkg') ++
+          showStanzas (stanzas pkg') ++ " " ++
+          case pr of
+            NewPackage     -> "(new package)"
+            NewVersion _   -> "(new version)"
+            Reinstall _ cs -> "(reinstall)" ++ case cs of
+                []   -> ""
+                diff -> " changes: "  ++ intercalate ", " (map change diff)
+
+    toFlagAssignment :: [Flag] -> FlagAssignment
+    toFlagAssignment = map (\ f -> (flagName f, flagDefault f))
+
+    nonDefaultFlags :: ConfiguredPackage -> FlagAssignment
+    nonDefaultFlags (ConfiguredPackage spkg fa _ _) =
+      let defaultAssignment =
+            toFlagAssignment
+             (genPackageFlags (Source.packageDescription spkg))
+      in  fa \\ defaultAssignment
+
+    stanzas :: ConfiguredPackage -> [OptionalStanza]
+    stanzas (ConfiguredPackage _ _ sts _) = sts
+
+    showStanzas :: [OptionalStanza] -> String
+    showStanzas = concatMap ((' ' :) . showStanza)
+    showStanza TestStanzas  = "*test"
+    showStanza BenchStanzas = "*bench"
+
+    -- FIXME: this should be a proper function in a proper place
+    showFlagAssignment :: FlagAssignment -> String
+    showFlagAssignment = concatMap ((' ' :) . showFlagValue)
+    showFlagValue (f, True)   = '+' : showFlagName f
+    showFlagValue (f, False)  = '-' : showFlagName f
+    showFlagName (FlagName f) = f
+
     change (OnlyInLeft pkgid)        = display pkgid ++ " removed"
     change (InBoth     pkgid pkgid') = display pkgid ++ " -> "
                                     ++ display (packageVersion pkgid')
     change (OnlyInRight      pkgid') = display pkgid' ++ " added"
-    changed (InBoth    pkgid pkgid') = pkgid /= pkgid'
-    changed _                        = True
 
 -- ------------------------------------------------------------
 -- * Post installation stuff
@@ -369,7 +539,7 @@ postInstallActions :: Verbosity
                    -> InstallPlan
                    -> IO ()
 postInstallActions verbosity
-  (packageDBs, _, comp, conf, globalFlags, configFlags, _, installFlags)
+  (packageDBs, _, comp, conf, globalFlags, configFlags, _, installFlags, _)
   targets installPlan = do
 
   unless oneShot $
@@ -458,8 +628,8 @@ regenerateHaddockIndex verbosity packageDBs comp conf
      "Updating documentation index " ++ indexFile
 
   --TODO: might be nice if the install plan gave us the new InstalledPackageInfo
-  installed <- getInstalledPackages verbosity comp packageDBs conf
-  Haddock.regenerateHaddockIndex verbosity installed conf indexFile
+  installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
+  Haddock.regenerateHaddockIndex verbosity installedPkgIndex conf indexFile
 
   | otherwise = return ()
   where
@@ -542,6 +712,8 @@ printBuildFailures plan =
                         ++ " The exception was:\n  " ++ show e
       BuildFailed     e -> " failed during the building phase."
                         ++ " The exception was:\n  " ++ show e
+      TestsFailed     e -> " failed during the tests phase."
+                        ++ " The exception was:\n  " ++ show e
       InstallFailed   e -> " failed during the final install step."
                         ++ " The exception was:\n  " ++ show e
 
@@ -557,21 +729,22 @@ data InstallMisc = InstallMisc {
 
 performInstallations :: Verbosity
                      -> InstallContext
-                     -> PackageIndex InstalledPackage
+                     -> PackageIndex
                      -> InstallPlan
                      -> IO InstallPlan
 performInstallations verbosity
   (packageDBs, _, comp, conf,
-   globalFlags, configFlags, configExFlags, installFlags)
-  installed installPlan = do
+   globalFlags, configFlags, configExFlags, installFlags, haddockFlags)
+  installedPkgIndex installPlan = do
 
   executeInstallPlan installPlan $ \cpkg ->
     installConfiguredPackage platform compid configFlags
                              cpkg $ \configFlags' src pkg ->
-      fetchAvailablePackage verbosity src $ \src' ->
+      fetchSourcePackage verbosity src $ \src' ->
         installLocalPackage verbosity (packageId pkg) src' $ \mpath ->
-          installUnpackedPackage verbosity (setupScriptOptions installed)
-                                 miscOptions configFlags' installFlags
+          installUnpackedPackage verbosity
+                                 (setupScriptOptions installedPkgIndex)
+                                 miscOptions configFlags' installFlags haddockFlags
                                  compid pkg mpath useLogFile
 
   where
@@ -646,7 +819,7 @@ executeInstallPlan plan installPkg = case InstallPlan.ready plan of
         -- which kind of means it was not their fault.
 
 
--- | Call an installer for an 'AvailablePackage' but override the configure
+-- | Call an installer for an 'SourcePackage' but override the configure
 -- flags with the ones given by the 'ConfiguredPackage'. In particular the
 -- 'ConfiguredPackage' specifies an exact 'FlagAssignment' and exactly
 -- versioned package dependencies. So we ignore any previous partial flag
@@ -658,24 +831,26 @@ installConfiguredPackage :: Platform -> CompilerId
                                          -> PackageDescription -> a)
                          -> a
 installConfiguredPackage platform comp configFlags
-  (ConfiguredPackage (AvailablePackage _ gpkg source) flags deps)
+  (ConfiguredPackage (SourcePackage _ gpkg source) flags stanzas deps)
   installPkg = installPkg configFlags {
     configConfigurationsFlags = flags,
-    configConstraints = map thisPackageVersion deps
+    configConstraints = map thisPackageVersion deps,
+    configBenchmarks = toFlag False,
+    configTests = toFlag (TestStanzas `elem` stanzas)
   } source pkg
   where
     pkg = case finalizePackageDescription flags
            (const True)
-           platform comp [] gpkg of
+           platform comp [] (enableStanzas stanzas gpkg) of
       Left _ -> error "finalizePackageDescription ConfiguredPackage failed"
       Right (desc, _) -> desc
 
-fetchAvailablePackage
+fetchSourcePackage
   :: Verbosity
   -> PackageLocation (Maybe FilePath)
   -> (PackageLocation FilePath -> IO BuildResult)
   -> IO BuildResult
-fetchAvailablePackage verbosity src installPkg = do
+fetchSourcePackage verbosity src installPkg = do
   fetched <- checkFetched src
   case fetched of
     Just src' -> installPkg src'
@@ -728,13 +903,14 @@ installUnpackedPackage :: Verbosity
                    -> InstallMisc
                    -> ConfigFlags
                    -> InstallFlags
+                   -> HaddockFlags
                    -> CompilerId
                    -> PackageDescription
                    -> Maybe FilePath -- ^ Directory to change to before starting the installation.
                    -> Maybe (PackageIdentifier -> FilePath) -- ^ File to log output to (if any)
                    -> IO BuildResult
 installUnpackedPackage verbosity scriptOptions miscOptions
-                       configFlags installConfigFlags
+                       configFlags installConfigFlags haddockFlags
                        compid pkg workingDir useLogFile =
 
   -- Configure phase
@@ -747,22 +923,27 @@ installUnpackedPackage verbosity scriptOptions miscOptions
 
   -- Doc generation phase
       docsResult <- if shouldHaddock
-        then (do setup haddockCommand haddockFlags
+        then (do setup haddockCommand haddockFlags'
                  return DocsOk)
                `catchIO`   (\_ -> return DocsFailed)
                `catchExit` (\_ -> return DocsFailed)
         else return DocsNotTried
 
   -- Tests phase
-      testsResult <- return TestsNotTried  --TODO: add optional tests
+      onFailure TestsFailed $ do
+        when (testsEnabled && PackageDescription.hasTests pkg) $
+            setup Cabal.testCommand testFlags
 
-  -- Install phase
-      onFailure InstallFailed $
-        withWin32SelfUpgrade verbosity configFlags compid pkg $ do
-          case rootCmd miscOptions of
-            (Just cmd) -> reexec cmd
-            Nothing    -> setup Cabal.installCommand installFlags
-          return (Right (BuildOk docsResult testsResult))
+        let testsResult | testsEnabled = TestsOk
+                        | otherwise = TestsNotTried
+
+      -- Install phase
+        onFailure InstallFailed $
+          withWin32SelfUpgrade verbosity configFlags compid pkg $ do
+            case rootCmd miscOptions of
+              (Just cmd) -> reexec cmd
+              Nothing    -> setup Cabal.installCommand installFlags
+            return (Right (BuildOk docsResult testsResult))
 
   where
     configureFlags   = filterConfigureFlags configFlags {
@@ -774,10 +955,11 @@ installUnpackedPackage verbosity scriptOptions miscOptions
       buildVerbosity = toFlag verbosity'
     }
     shouldHaddock    = fromFlag (installDocumentation installConfigFlags)
-    haddockFlags _   = emptyHaddockFlags {
-      haddockDistPref  = configDistPref configFlags,
+    haddockFlags' _   = haddockFlags {
       haddockVerbosity = toFlag verbosity'
     }
+    testsEnabled = fromFlag (configTests configFlags)
+    testFlags _ = Cabal.emptyTestFlags
     installFlags _   = Cabal.emptyInstallFlags {
       Cabal.installDistPref  = configDistPref configFlags,
       Cabal.installVerbosity = toFlag verbosity'
